function[u,v,w] = ConstrainedGust(x,y,z,A,x0,l,shape,seed,spectrum,varargin)
% Constrained stochastic simulation of a spatial wind gust. 
% 
% 	[u v w] = ConstrainedGust(x,y,z,A,x0,l,shape,seed,spectrum,varargin)
%     
% 	Returns a three-dimensional turbulent velocity field with a wind gust
% 	embedded into it. The resulting field is obtained by constraining the
% 	randomization of the wave numbers, yielding a stochastic velocity field 
%   that satisfies a certain input while adhering to the statistics.
%     
% 	INPUT:
%	x,y,z    (double)   Position vectors as created by ndgrid [m].
%	A        (double)   Gust amplitude averaged over volume.
%	x0       (double)   Three-element vector containing the [x,y,z]-
%                       position of the of gust center [m].
%	l        (double)   Length scale of the gust. This is either a three-
%                       element vector containing the dimensions of the 
%                       volume along the [x,y,z]-axes, or a single value if
%                       all lengths are equal.
%	shape    (string)   Gust shape ('Box'/'Ellipsoid').
%   seed     (str/dbl)  Turbulence seed used by Matlab's rng() function.
%                       Either a seed number, 'Shuffle' for an arbitrary
%                       seed, or 'Mean' to return the gust's mean shape.
%	spectrum (string)   Spectrum name ('Isotropic'/'Kaimal'/'KaimalIEC'/
%                       'Mann','MannIEC')
%   varargin            Spectrum parameters (see below).
%
%       Isotropic:
%       [u,v,w] = ConstrainedGust(~,'Isotropic',sigma,L)
%       sigma   (double)    Isotropic standard deviation [m/s].
%       L       (double)    Turbulent length scale [m].
%
%       Kaimal:
%       [u,v,w] = ConstrainedGust(~,'Kaimal',U,sigma1,sigma2,sigma3,...
%                                                               Lu,Lv,Lw,H)
%       U       (double)    Mean wind speed [m/s].
%       sigma1  (double)    Longitudinal standard deviation [m/s].
%       sigma2  (double)    Lateral standard deviation [m/s].
%       sigma3  (double)    Vertical standard deviation [m/s].
%       Lu      (double)    Longitudinal integral scale parameter [m].
%       Lv      (double)    Lateral integral scale parameter [m].
%       Lw      (double)    Vertical integral scale parameter [m].
%       H       (double)    Exponential decay parameter [-] (= 12 for IEC).
%
%       KaimalIEC:
%       [u,v,w] = ConstrainedGust(~,'KaimalIEC',U,sigma1,h)
%       U       (double)    Mean wind speed [m/s].
%       sigma1  (double)    Longitudinal standard deviation [m/s].
%       h       (double)    Hub height [m],
%
%       Mann:
%       [u,v,w] = ConstrainedGust(~,'Mann',Gamma,L,ae)
%       Gamma   (double)    Anisotropy parameter [-] (= 0 for isotropic
%                           turbulence).
%       L       (double)    Turbulent length scale [m].
%       ae      (double)    Coefficient of the inertial cascade, a*e^(2/3), 
%                           where a = 1.7 is the three-dimensional
%                           Kolmogorov constant [-] and e = the mean TKE 
%                           dissipation rate [m/s].
%
%       MannIEC:
%       [u,v,w] = ConstrainedGust(~,'MannIEC',sigma1,h)
%       sigma1  (double)    Longitudinal standard deviation [m/s].
%       h       (double)    Hub height [m],
%
%	OUTPUT:
%	u,v,w   (double)    Velocity components [m/s].
%     
% 	AUTHOR: 
%   Ren Bos
% 	R.Bos-1@tudelft.nl
% 	Delft University of Technology
% 	Last updated March 2017
%
%   LITERATURE:
%   - Bos, Ren (2017). "Extreme gusts and their role in wind turbine
%     design." PhD thesis. Delft University of Technology.
%   - Bierbooms, Wim (2005). "Constrained Stochastic Simulation -- 
%     Generation of Time Series around Some Specific Event in a Normal 
%     Process." Extremes 8 (3): pp. 207224.
%   - IEC (2005). IEC 61400-1 Wind Turbines - Part 1: Design Requirements, 
%     3rd ed. International Electrotechnical Commission.
%   - Mann, Jakob (1998). "Wind Field Simulation." Probabilistic 
%     Engineering Mechanics 13(4), pp. 269282.

%% Domain size   
Lx = abs(x(end) - x(1));
Ly = abs(y(end) - y(1));
Lz = abs(z(end) - z(1));

Nx = size(x,1);
Ny = size(y,2);
Nz = size(z,3);

%% Wave number discretization
[mx, my, mz] = ndgrid(...
    -Nx/2:Nx/2-1, ...
    -Ny/2:Ny/2-1, ...
    -Nz/2:Nz/2-1);

mx = ifftshift(mx);
my = ifftshift(my);
mz = ifftshift(mz);

kx = 2*pi*mx/Lx;
ky = 2*pi*my/Ly;
kz = 2*pi*mz/Lz;

k = sqrt(kx.^2 + ky.^2 + kz.^2);

%% Turbulence models
% Isotropic
if strcmpi(spectrum, 'Isotropic')
    
    % Input parameters
    sigma_iso = varargin{1};
    L = varargin{2};
    
    % Von Krmn isotropic turbulence spectrum
    E = 4*gamma(17/6)/(sqrt(pi)*gamma(1/3)) * sigma_iso^2*L .* (k*L).^4 ./ (1 + (k*L).^2).^(17/6);

    % Correlation tensor
    B = sqrt(2*pi^2 * E ./ (Lx*Ly*Lz * k.^4));
    C = zeros([3,3,size(k)]);
    C(1,2,:,:,:) = B .* kz;
    C(1,3,:,:,:) = B .* -ky;
    C(2,1,:,:,:) = B .* -kz;
    C(2,3,:,:,:) = B .* kx;
    C(3,1,:,:,:) = B .* ky;
    C(3,2,:,:,:) = B .* -kx;
    C(:,:,1,:,:) = 0;

% Kaimal spectrum
elseif strcmpi(spectrum, 'Kaimal')
    
    % Input parameters
    U = varargin{1};
    sigma1 = varargin{2};
    sigma2 = varargin{3};
    sigma3 = varargin{4};
    Lu = varargin{5};
    Lv = varargin{6};
    Lw = varargin{7};
    H = varargin{8};    
   
    % Single-point spectra
    Fu = 1/2 * sigma1^2 * 4*Lu/U./(1+6/(2*pi)*abs(kx)*Lu).^(5/3);
    Fv = 1/2 * sigma2^2 * 4*Lv/U./(1+6/(2*pi)*abs(kx)*Lv).^(5/3);
    Fw = 1/2 * sigma3^2 * 4*Lw/U./(1+6/(2*pi)*abs(kx)*Lw).^(5/3);
    
    % Coherence function
    dy = y(1,2,1) - y(1,1,1);
    dz = z(1,1,2) - z(1,1,1);
    r = sqrt((my*dy).^2 + (mz*dz).^2);
    Coh_u = exp(-H*r.*sqrt((kx/(2*pi)).^2 + (0.12/Lu)^2));
    Coh_v = exp(-H*r.*kx/(2*pi));
    Coh_w = exp(-H*r.*kx/(2*pi));
    
    % Correlation matrix
    C = zeros([3,3,size(k)]);
    C(1,1,:,:,:) = sqrt((2*pi)^3/(Lx*Ly*Lz) * Fu.* fft(fft(Coh_u,[],2),[],3));
    C(2,2,:,:,:) = sqrt((2*pi)^3/(Lx*Ly*Lz) * Fv.* fft(fft(Coh_v,[],2),[],3));
    C(3,3,:,:,:) = sqrt((2*pi)^3/(Lx*Ly*Lz) * Fw.* fft(fft(Coh_w,[],2),[],3));
    
    % Empty matrix along kx = 0
    C(:,:,1,:,:) = 0;
    
% Kaimal spectrum (IEC 61400-1)
elseif strcmpi(spectrum, 'KaimalIEC')
    
    % Input parameters
    U = varargin{1};
    sigma1 = varargin{2};
    sigma2 = 0.8*sigma1;
    sigma3 = 0.5*sigma1;
    h = varargin{3};
    if h < 60
        Lambda = 0.7*h;
    else
        Lambda = 42;
    end 
    Lu = 8.1*Lambda;
    Lv = 2.7*Lambda;
    Lw = 0.66*Lambda;
    
    % Single-point spectra
    Fu = 1/2 * sigma1^2 * 4*Lu/U./(1+6/(2*pi)*abs(kx)*Lu).^(5/3);
    Fv = 1/2 * sigma2^2 * 4*Lv/U./(1+6/(2*pi)*abs(kx)*Lv).^(5/3);
    Fw = 1/2 * sigma3^2 * 4*Lw/U./(1+6/(2*pi)*abs(kx)*Lw).^(5/3);
    
    % Coherence function
    dy = y(1,2,1) - y(1,1,1);
    dz = z(1,1,2) - z(1,1,1);
    r = sqrt((my*dy).^2 + (mz*dz).^2);
    Coh_u = exp(-12*sqrt((kx.*r/(2*pi)).^2 + (0.12*r/Lu).^2));
    
    % Correlation matrix
    C = zeros([3,3,size(k)]);
    C(1,1,:,:,:) = sqrt((2*pi)^3/(Lx*Ly*Lz) * Fu.* fft(fft(Coh_u,[],2),[],3));
    C(2,2,:,:,:) = sqrt((2*pi)^3/(Lx*Ly*Lz) * Fv);
    C(3,3,:,:,:) = sqrt((2*pi)^3/(Lx*Ly*Lz) * Fw);
    
    % Empty matrix along kx = 0
    C(:,:,1,:,:) = 0;
    
% Mann spectrum
elseif strcmpi(spectrum, 'Mann')
    
    % Input parameters
    Gamma = varargin{1};
    L = varargin{2};
    ae = varargin{3};
    
    % Non-dimensional distortion time
    E = @(p) 4*gamma(17/6)/(sqrt(pi)*gamma(1/3))*p.^4./(1+p.^2).^(17/6);
    k_ = linspace(0,max(abs(k(:)))*L,1e3+1);
    E_ = zeros(1,length(k_));
    E_(1) = 1.5;
    for i = 1:1e3
    	E_(i+1) = E_(i) - integral(E,k_(i),k_(i+1));
    end
    if Gamma > 0
    	beta = 1.476*Gamma ./ (k*L .* sqrt(interp1(k_,E_,k*L,'pchip')));  
    	beta(1,1,1) = 1.476*Gamma ./ (0.1*k(1,1,2)*L .* sqrt(interp1(k_,E_,0.1*k(1,1,2)*L,'pchip')));  
    else
    	beta = zeros(Nx,Ny,Nz);
    end

    % Undistorted wave number
    kz0 = kz + beta.*kx;
    k0 = sqrt(kx.^2 + ky.^2 + kz0.^2);

    % Von Krmn isotropic turbulence spectrum
    E0 = ae * L^(5/3) * (L*k0).^4 ./ (1 + (L*k0).^2).^(17/6);

    % Shear model parameters
    C1 = (beta.*kx.^2 .* (kx.^2 + ky.^2 - kz.*(kz + beta.*kx))) ./ (k.^2 .* (kx.^2 + ky.^2));
    C2 = (ky.*k0.^2) ./ (kx.^2 + ky.^2).^(3/2) .* atan2(real(beta.*kx.*sqrt(kx.^2 + ky.^2)), real(k0.^2 - (kz + beta.*kx).*kx.*beta));
    zeta1 = C1 - ky./kx.*C2;
    zeta2 = C2 + ky./kx.*C1;

    % Spectral tensor
    Phi = zeros([3,3,size(k)]);
    Phi(1,1,:,:,:) = E0./(4*pi*k0.^4) .* (k0.^2 - kx.^2 - 2*kx.*kz0.*zeta1 + (kx.^2+ky.^2).*zeta1.^2);
    Phi(1,2,:,:,:) = E0./(4*pi*k0.^4) .* (-kx.*ky - kx.*kz0.*zeta2 - ky.*kz0.*zeta1 + (kx.^2+ky.^2).*zeta1.*zeta2);
    Phi(1,3,:,:,:) = E0./(4*pi*k0.^2.*k.^2) .* (-kx.*kz0 + (kx.^2+ky.^2).*zeta1);
    Phi(2,1,:,:,:) = conj(Phi(1,2,:,:,:));
    Phi(2,2,:,:,:) = E0./(4*pi*k0.^4) .* (k0.^2 - ky.^2 - 2*ky.*kz0.*zeta2 + (kx.^2+ky.^2).*zeta2.^2);
    Phi(2,3,:,:,:) = E0./(4*pi*k0.^2.*k.^2) .* (-ky.*kz0 + (kx.^2+ky.^2).*zeta2);
    Phi(3,1,:,:,:) = conj(Phi(1,3,:,:,:));
    Phi(3,2,:,:,:) = conj(Phi(2,3,:,:,:));
    Phi(3,3,:,:,:) = E0./(4*pi*k.^4) .* (kx.^2 + ky.^2);
    Phi(isnan(Phi)) = 0;

    % Correlation tensor
    C = zeros([3,3,size(k)]);
    for ix = 1:Nx
        for iy= 1:Ny
            for iz = 1:Nz
                [Y,a] = eig(Phi(:,:,ix,iy,iz));
                C(:,:,ix,iy,iz) = real(Y*sqrt(a) * sqrt((2*pi)^3/(Lx*Ly*Lz)));
            end
        end
    end
    
% Mann spectrum (IEC 61400-1)
elseif strcmpi(spectrum, 'MannIEC')
    
    % Input parameters
    sigma1 = varargin{1};
    sigma_iso = 0.55*sigma1;
    Gamma = 3.9;
    h = varargin{2};
    if h < 60
        Lambda = 0.7*h;
    else
        Lambda = 42;
    end 
    L = 0.8*Lambda;
    
    % Non-dimensional distortion time
    E = @(p) 4*gamma(17/6)/(sqrt(pi)*gamma(1/3))*p.^4./(1+p.^2).^(17/6);
    k_ = linspace(0,max(abs(k(:)))*L,1e3+1);
    E_ = zeros(1,length(k_));
    E_(1) = 1.5;
    for i = 1:1e3
    	E_(i+1) = E_(i) - integral(E,k_(i),k_(i+1));
    end
    if Gamma > 0
    	beta = 1.476*Gamma ./ (k*L .* sqrt(interp1(k_,E_,k*L,'pchip')));  
    	beta(1,1,1) = 1.476*Gamma ./ (0.1*k(1,1,2)*L .* sqrt(interp1(k_,E_,0.1*k(1,1,2)*L,'pchip')));  
    else
    	beta = zeros(Nx,Ny,Nz);
    end

    % Undistorted wave number
    kz0 = kz + beta.*kx;
    k0 = sqrt(kx.^2 + ky.^2 + kz0.^2);

    % Von Krmn isotropic turbulence spectrum
    E0 = sigma_iso^2 * 4*gamma(17/6)/(sqrt(pi)*gamma(1/3)) * L * (k0*L).^4 ./ (1 + (k0*L).^2).^(17/6);

    % Shear model parameters
    C1 = (beta.*kx.^2 .* (kx.^2 + ky.^2 - kz.*(kz + beta.*kx))) ./ (k.^2 .* (kx.^2 + ky.^2));
    C2 = (ky.*k0.^2) ./ (kx.^2 + ky.^2).^(3/2) .* atan2(real(beta.*kx.*sqrt(kx.^2 + ky.^2)), real(k0.^2 - (kz + beta.*kx).*kx.*beta));
    zeta1 = C1 - ky./kx.*C2;
    zeta2 = C2 + ky./kx.*C1;
    
    % Spectral tensor
    Phi = zeros([3,3,size(k)]);
    Phi(1,1,:,:,:) = E0./(4*pi*k0.^4) .* (k0.^2 - kx.^2 - 2*kx.*kz0.*zeta1 + (kx.^2+ky.^2).*zeta1.^2);
    Phi(1,2,:,:,:) = E0./(4*pi*k0.^4) .* (-kx.*ky - kx.*kz0.*zeta2 - ky.*kz0.*zeta1 + (kx.^2+ky.^2).*zeta1.*zeta2);
    Phi(1,3,:,:,:) = E0./(4*pi*k0.^2.*k.^2) .* (-kx.*kz0 + (kx.^2+ky.^2).*zeta1);
    Phi(2,1,:,:,:) = conj(Phi(1,2,:,:,:));
    Phi(2,2,:,:,:) = E0./(4*pi*k0.^4) .* (k0.^2 - ky.^2 - 2*ky.*kz0.*zeta2 + (kx.^2+ky.^2).*zeta2.^2);
    Phi(2,3,:,:,:) = E0./(4*pi*k0.^2.*k.^2) .* (-ky.*kz0 + (kx.^2+ky.^2).*zeta2);
    Phi(3,1,:,:,:) = conj(Phi(1,3,:,:,:));
    Phi(3,2,:,:,:) = conj(Phi(2,3,:,:,:));
    Phi(3,3,:,:,:) = E0./(4*pi*k.^4) .* (kx.^2 + ky.^2);
    Phi(isnan(Phi)) = 0;

    % Correlation tensor
    C = zeros([3,3,size(k)]);
    for ix = 1:Nx
        for iy= 1:Ny
            for iz = 1:Nz
                [Y,a] = eig(Phi(:,:,ix,iy,iz));
                C(:,:,ix,iy,iz) = real(Y*sqrt(a) * sqrt((2*pi)^3/(Lx*Ly*Lz)));
            end
        end
    end
    
end
    
%% Apply constraints
% Constraint vector
b = [A; 0; 0; 0];
x0 = x0(:);

% Low-pass filter
if length(l) == 1
    lx = l;
    ly = l;
    lz = l;
elseif length(l) == 3
    lx = l(1);
    ly = l(2);
    lz = l(3);
end
if strcmpi(shape, 'Box')
    dx = Lx/(Nx-1);
    dy = Ly/(Ny-1);
    dz = Lz/(Nz-1);
    G = sinc(mx*lx/(Nx*dx)) .* sinc(my*ly/(Ny*dy)) .* sinc(mz*lz/(Nz*dz));
elseif strcmpi(shape, 'Ellipsoid')
    if lx == 0 && ly == 0 && lz == 0
        G = ones(size(kx));
    else
        K = sqrt((kx*lx/2).^2 + (ky*ly/2).^2 + (kz*lz/2).^2);
        G = 3./K.^3 .* (sin(K) - K.*cos(K));
        G(K==0) = 1;
    end
end

% Reshape wave numbers to vector
kx = reshape(kx,1,[]);
ky = reshape(ky,1,[]);
kz = reshape(kz,1,[]);
C = reshape(C,3,[]);
G = reshape(G,1,[]);

% DFT matrix with first derivatives
Y = [C(1,:) .* kron(G .* exp(1i * (kx*(x0(1)-x(1)) + ky*(x0(2)-y(1)) + kz*(x0(3)-z(1)))), ones(1,3)); ...
     C(1,:) .* kron(G .* 1i.*kx .* exp(1i * (kx*(x0(1)-x(1)) + ky*(x0(2)-y(1)) + kz*(x0(3)-z(1)))), ones(1,3)); ...
     C(1,:) .* kron(G .* 1i.*ky .* exp(1i * (kx*(x0(1)-x(1)) + ky*(x0(2)-y(1)) + kz*(x0(3)-z(1)))), ones(1,3)); ...
     C(1,:) .* kron(G .* 1i.*kz .* exp(1i * (kx*(x0(1)-x(1)) + ky*(x0(2)-y(1)) + kz*(x0(3)-z(1)))), ones(1,3))];
 
% Constrained stochastic variable
if strcmpi(seed, 'Mean')
    nc = Y'/(Y*Y') * b;
else
    rng(seed);
    n = randn(size(Y,2),1) + 1i*randn(size(Y,2),1);
    nc = n + Y'/(Y*Y') * (b - Y*n);
end

%% Stochastic field
dZ = zeros([3, size(k)]);
nc = reshape(nc, [3, size(k)]);
C  = reshape(C,  [3, 3, size(k)]);
dZ(1,:,:,:) = sum(squeeze(C(1,:,:,:,:)).*nc,1);
dZ(2,:,:,:) = sum(squeeze(C(2,:,:,:,:)).*nc,1);
dZ(3,:,:,:) = sum(squeeze(C(3,:,:,:,:)).*nc,1);

%% Reconstruct time series
u = Nx*Ny*Nz*real(ifftn(squeeze(dZ(1,:,:,:))));
v = Nx*Ny*Nz*real(ifftn(squeeze(dZ(2,:,:,:))));
w = Nx*Ny*Nz*real(ifftn(squeeze(dZ(3,:,:,:))));
