
clear all
close all

% What follows is a series of examples to demonstrate the ConstrainedGust
% script. Please note that, since the variances are not corrected, the gust
% amplitude and standard deviations do not always match the prescribed 
% values perfectly. Increasing the number of grid points will usually 
% improve the quality of the result. Also, the Mann spectrum may take a
% while to set up due to the Cholesky decomposition. If a large number of
% such gusts are required, precomputing the correlation tensors will speed
% up the process significantly.

%% Example 1: Mean shape of point gust in isotropic turbulence
% Domain
[x,y,z] = ndgrid(...
    linspace(0,300,2^6), ...
    linspace(0,300,2^6), ...
    linspace(0,300,2^6));

% Gust parameters
A = 6;              % Gust amplitude [m/s]
l = 0;              % Gust volume [m]
shape = 'Box';      % Gust shape ('Box'/'Ellipsoid'). 
x0 = [x(20,1,1);... % Gust position [m]
      y(1,20,1);...
      z(1,1,20)];

% Spectrum
spectrum = 'Isotropic';
sigma_iso = 1;      % Isotropic standard deviation [m/s].
L = 25;             % Turbulent length scale [m].

% Turbulence seed
seed = 'Mean';

% Velocity field
[u,v,w] = ConstrainedGust(x,y,z,A,x0,l,shape,seed,spectrum,sigma_iso,L);

% Surface plot through z0
figure()
surf(squeeze(y(20,:,:)), squeeze(z(20,:,:)), squeeze(u(20,:,:)))
xlabel('y')
ylabel('z')
zlabel('u')

% Time series through y0,z0
figure()
hold on
plot(squeeze(x(:,20,20)), u(:,20,20),'r')
plot(squeeze(x(:,20,20)), v(:,20,20),'b')
plot(squeeze(x(:,20,20)), w(:,20,20),'m')
xlabel('x')
ylabel('u,v,w')
legend('u','v','w')

%% Example 2: Point gust in Kaimal (IEC) turbulence
% Domain
[x,y,z] = ndgrid(...
    linspace(0,300,2^6), ...
    linspace(0,300,2^6), ...
    linspace(0,300,2^6));

% Gust parameters
A = 6;              % Gust amplitude [m/s]
l = 0;              % Gust volume [m]
shape = 'Box';      % Gust shape ('Box'/'Ellipsoid'). 
x0 = [x(20,1,1);... % Gust position [m]
      y(1,20,1);...
      z(1,1,20)];

% Spectrum
spectrum = 'KaimalIEC';
U = 10;             % Mean wind speed [m/s].
sigma1 = 1;         % Longitudinal standard deviation [m/s].
h = 90;             % Hub height [m].

% Turbulence seed
seed = 'Shuffle';

% Velocity field
[u,v,w] = ConstrainedGust(x,y,z,A,x0,l,shape,seed,spectrum,U,sigma1,h);

% Surface plot through x0
figure()
surf(squeeze(y(20,:,:)), squeeze(z(20,:,:)), squeeze(u(20,:,:)))
xlabel('y')
ylabel('z')
zlabel('u')

% Time series through y0,z0
figure()
hold on
plot(squeeze(x(:,20,20)), u(:,20,20),'r')
plot(squeeze(x(:,20,20)), v(:,20,20),'b')
plot(squeeze(x(:,20,20)), w(:,20,20),'m')
xlabel('x')
ylabel('u,v,w')
legend('u','v','w')

%% Example 3: Spherical gust in Mann (IEC) turbulence
% Domain
[x,y,z] = ndgrid(...
    linspace(0,300,2^6), ...
    linspace(0,300,2^6), ...
    linspace(0,300,2^6));

% Gust parameters
A = 4;              % Gust amplitude [m/s]
l = 50;             % Gust volume [m]
shape = 'Ellipsoid';% Gust shape ('Box'/'Ellipsoid'). 
x0 = [x(20,1,1);... % Gust position [m]
      y(1,20,1);...
      z(1,1,20)];

% Spectrum
spectrum = 'MannIEC';
sigma1 = 1;         % Longitudinal standard deviation [m/s].
h = 90;             % Hub height [m].

% Turbulence seed
seed = 'Shuffle';

% Velocity field
[u,v,w] = ConstrainedGust(x,y,z,A,x0,l,shape,seed,spectrum,sigma1,h);

% Surface plot through x0
figure()
surf(squeeze(y(20,:,:)), squeeze(z(20,:,:)), squeeze(u(20,:,:)))
xlabel('y')
ylabel('z')
zlabel('u')

% Time series through y0,z0
figure()
hold on
plot(squeeze(x(:,20,20)), u(:,20,20),'r')
plot(squeeze(x(:,20,20)), v(:,20,20),'b')
plot(squeeze(x(:,20,20)), w(:,20,20),'m')
xlabel('x')
ylabel('u,v,w')
legend('u','v','w')
